﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using HeuristicLab.MainForm;
using HeuristicLab.PluginInfrastructure;
using PluginDeploymentService = HeuristicLab.PluginInfrastructure.Advanced.DeploymentService;
using HeuristicLab.PluginInfrastructure.Manager;
using System.ServiceModel;
using ICSharpCode.SharpZipLib.Zip;
using System.IO;

namespace HeuristicLab.DeploymentService.AdminClient {
  public partial class PluginListView : HeuristicLab.MainForm.WindowsForms.View {
    private Dictionary<IPluginDescription, PluginDeploymentService.PluginDescription> localAndServerPlugins;
    private BackgroundWorker pluginUploadWorker;
    private BackgroundWorker updateServerPluginsWorker;

    public PluginListView() {
      InitializeComponent();
      Caption = "Plugins";

      using (var client = new PluginDeploymentService.UpdateClient()) {
        serverUrlTextBox.Text = client.Endpoint.Address.ToString();
      }

      localAndServerPlugins = new Dictionary<IPluginDescription, PluginDeploymentService.PluginDescription>();

      pluginUploadWorker = new BackgroundWorker();
      pluginUploadWorker.DoWork += new DoWorkEventHandler(pluginUploadWorker_DoWork);
      pluginUploadWorker.RunWorkerCompleted += new RunWorkerCompletedEventHandler(pluginUploadWorker_RunWorkerCompleted);

      updateServerPluginsWorker = new BackgroundWorker();
      updateServerPluginsWorker.DoWork += new DoWorkEventHandler(updateServerPluginsWorker_DoWork);
      updateServerPluginsWorker.RunWorkerCompleted += new RunWorkerCompletedEventHandler(updateServerPluginsWorker_RunWorkerCompleted);
    }

    void updateServerPluginsWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e) {
      if (!e.Cancelled && e.Result != null) {
        UpdateLocalPlugins();
        var plugins = (PluginDeploymentService.PluginDescription[])e.Result;
        foreach (var plugin in plugins) {
          var matchingLocalPlugin = (from localPlugin in localAndServerPlugins.Keys
                                     where localPlugin.Name == plugin.Name
                                     where localPlugin.Version == localPlugin.Version
                                     select localPlugin).SingleOrDefault();
          if (matchingLocalPlugin != null) {
            localAndServerPlugins[matchingLocalPlugin] = plugin;
          }
        }
        UpdatePluginList();
        OnConnected();
      } else {
        OnDisconnected();
      }
      Cursor = Cursors.Default;
    }

    private void UpdatePluginList() {
      listView.Items.Clear();
      foreach (var pair in localAndServerPlugins) {
        var item = MakeListViewItem(pair.Key);
        listView.Items.Add(item);
      }
    }

    void updateServerPluginsWorker_DoWork(object sender, DoWorkEventArgs e) {
      try {
        using (var client = new PluginDeploymentService.UpdateClient()) {
          e.Result = client.GetPlugins();
          e.Cancel = false;
        }
      }
      catch (EndpointNotFoundException ex) {
        e.Result = null;
        e.Cancel = true;
      }
      catch (FaultException ex) {
        e.Result = null;
        e.Cancel = true;
      }
    }

    void pluginUploadWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e) {
      this.Enabled = true;
    }

    void pluginUploadWorker_DoWork(object sender, DoWorkEventArgs e) {
      var selectedPlugins = (IEnumerable<IPluginDescription>)e.Argument;
      using (PluginDeploymentService.AdminClient adminClient = new PluginDeploymentService.AdminClient()) {

        foreach (var plugin in IteratePlugins(selectedPlugins)) {
          adminClient.DeployPlugin(MakePluginDescription(plugin), CreateZipPackage(plugin));
        }
      }
    }

    private byte[] CreateZipPackage(IPluginDescription plugin) {
      using (MemoryStream stream = new MemoryStream()) {
        ZipFile zipFile = new ZipFile(stream);
        zipFile.BeginUpdate();
        foreach (var file in plugin.Files) {
          zipFile.Add(file.Name);
        }
        zipFile.CommitUpdate();
        stream.Seek(0, SeekOrigin.Begin);
        return stream.GetBuffer();
      }
    }

    private PluginDeploymentService.PluginDescription MakePluginDescription(IPluginDescription plugin) {
      var dependencies = from dep in plugin.Dependencies
                         select MakePluginDescription(dep);
      if (string.IsNullOrEmpty(plugin.ContactName) || string.IsNullOrEmpty(plugin.ContactEmail)) {
        return new PluginDeploymentService.PluginDescription(plugin.Name, plugin.Version, dependencies);
      } else return new PluginDeploymentService.PluginDescription(plugin.Name, plugin.Version, dependencies, plugin.ContactName + ", " + plugin.ContactEmail, plugin.LicenseText);
    }

    private IEnumerable<IPluginDescription> IteratePlugins(IEnumerable<IPluginDescription> plugins) {
      foreach (var plugin in plugins) {
        foreach (var dependency in IteratePlugins(plugin.Dependencies)) {
          yield return dependency;
        }
        yield return plugin;
      }
    }

    public void UpdateLocalPlugins() {
      localAndServerPlugins.Clear();

      foreach (var plugin in ApplicationManager.Manager.Plugins) {
        localAndServerPlugins.Add(plugin, null);
      }
    }


    private ListViewItem MakeListViewItem(IPluginDescription plugin) {
      ListViewItem item;
      if (localAndServerPlugins[plugin] != null) {
        item = new ListViewItem(new string[] { plugin.Name, plugin.Version.ToString(), localAndServerPlugins[plugin].Version.ToString() });
      } else {
        item = new ListViewItem(new string[] { plugin.Name, plugin.Version.ToString(), string.Empty });
      }
      item.Tag = plugin;
      return item;
    }

    private void listView_SelectedIndexChanged(object sender, EventArgs e) {
      localPluginDetailsPanel.Controls.Clear();
      serverPluginDetailsGroupBox.Controls.Clear();

      if (listView.SelectedItems.Count > 0) {
        var plugin = (IPluginDescription)listView.SelectedItems[0].Tag;
        ShowPluginDetails(plugin);
      }
    }

    private void ShowPluginDetails(IPluginDescription plugin) {
      var localView = new LocalPluginView(plugin);
      var serverView = new ServerPluginView(localAndServerPlugins[plugin]);
      localView.Dock = DockStyle.Fill;
      serverView.Dock = DockStyle.Fill;
      localPluginDetailsPanel.Controls.Add(localView);
      serverPluginDetailsGroupBox.Controls.Add(serverView);
    }

    private void uploadButton_Click(object sender, EventArgs e) {
      if (listView.SelectedItems.Count > 0) {
        var selectedPlugins = from item in listView.SelectedItems.Cast<ListViewItem>()
                              where item.Tag is IPluginDescription
                              select item.Tag as IPluginDescription;
        this.Enabled = false;
        pluginUploadWorker.RunWorkerAsync(selectedPlugins.ToList());
      }
    }

    private void connectButton_Click(object sender, EventArgs e) {
      OnDisconnected();
      Cursor = Cursors.AppStarting;
      updateServerPluginsWorker.RunWorkerAsync();
    }

    private void OnConnected() {
      connectButton.Enabled = false;
      serverUrlTextBox.Enabled = false;
      label1.Enabled = false;

      listView.Enabled = true;
      uploadButton.Enabled = true;
    }

    private void OnDisconnected() {
      connectButton.Enabled = true;
      serverUrlTextBox.Enabled = true;
      label1.Enabled = true;

      localAndServerPlugins.Clear();
      listView.Items.Clear();
      localPluginDetailsPanel.Controls.Clear();
      serverPluginDetailsGroupBox.Controls.Clear();
      listView.Enabled = false;
      uploadButton.Enabled = false;
    }
  }
}
